"""NDG SOAP server package

NERC DataGrid Project
"""
__author__ = "Philip Kershaw"
__date__ = "26/01/2010"
__copyright__ = ""
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: $'
