"""SAML 2.0 Binding package

Implementation of SAML 2.0 for NDG Security

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "30/06/10"
__copyright__ = "(C) 2009 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: __init__.py 7130 2010-06-30 13:33:07Z pjkersha $"
