"""NDG SAML SOAP Binding unit test package

NERC DataGrid Project
"""
__author__ = "P J Kershaw"
__date__ = "30/06/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__license__ = "http://www.apache.org/licenses/LICENSE-2.0"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: __init__.py 7154 2010-07-01 15:59:59Z pjkersha $'
